classdef Fireworks < handle
    % Fireworks     Class for creating fireworks objects
    %
    % This class creates Fireworks objects that can be launched (and drawn)
    % on an axes. The fireworks come in different shapes, sizes, speeds,
    % and position. They also produce sounds.
    %
    % Fireworks Properties:
    %   FireworksType   - Type of fireworks
    %   NumPetals       - Number of petals after explosion
    %   SymmetricPetals - Whether the petals spread symmetrically
    %   LaunchPos       - Horizontal launch location
    %   LaunchVel       - Launch velocity (horizontal and vertical)
    %   SpeedFactor     - Speed of the fireworks
    %   HaveSound       - Whether there is sound
    %   ThinLines       - Whether the fireworks are displayed with thin lines
    %
    %   ValidFireworksType (Constant) - Valid input string for FireworksType
    %
    % Fireworks Methods:
    %   Fireworks       - Constructor
    %   step            - Step fireworks animation
    %   flyBlocking     - Fly fireworks to completion (blocking mode)
    %   newFireworks    - Change fireworks properties
    %   resetFireworks  - Reset animation to intial condition
    %
    % When you create a Fireworks object, it becomes tied to an axes (you
    % specify that as the first argument of the constructor as seen below).
    % Each Fireworks object has default random values for the properties.
    % They can be changed after construction, or you can specify them
    % during construction. See the help for the constructor.
    %
    % When an axes gets tied to the fireworks object(s), some of the axes
    % properties will be changed:
    %   * Background becomes black
    %   * All tick labels go away
    %   * Y limits become [0 3000]
    %   * Data aspect ratio becomes 1 to 1
    %   * X limits are automatically chosen based to the last two bullet
    %     points. The origin will be at the bottom center.
    %   * Internal appdata is added
    %
    % Use the "step" method to step through the animation. Note that the
    % animation is time-based, not frame based. For example, if you call
    % "step" for the first time and wait too long, the animation will be
    % over the next time you call "step". So in practice, the call to
    % "step" should be in a loop or run with a timer.
    %
    % Alternatively, you can call "flyBlocking" method which would animate
    % the fireworks in its entirety. This is a blocking call, so it does
    % not return control until the fireworks have completed animating.
    %
    % "resetFireworks" will reset the fireworks to the initial state
    % (launch time). Use this to animate the same fireworks one more time.
    % Note that "flyBlocking" calls "resetFireworks" beforehand, so there
    % is not need to call it separately.
    %
    % Use "newFireworks" to change the properties of the fireworks object.
    % Using this method (as opposed to changing properties one by one)
    % allows you to change multiple properties at once.
    %
    % Examples:
    %   % Create one fireworks object and animate the whole flight.
    %      f = Fireworks(gca)
    %      flyBlocking(f);
    %
    %   % Create 10 fireworks launching from the origin. The other
    %   % properties are randomly set.
    %      f2 = Fireworks(gca, 10, 'LaunchPos', 0);
    %      flyBlocking(f2);
    %
    %   % Change the FireworksType property of all 10 fireworks to
    %   % 'matlab'. The other properties are randomly set.
    %      newFireworks(f2, 'FireworksType', 'matlab');
    %      flyBlocking(f2);
    %
    % See also Fireworks, Fireworks.step, Fireworks.flyBlocking,
    % Fireworks.newFireworks, Fireworks.resetFireworks, Sky.
    %
    % For more info, <a href="matlab:doc('Fireworks')">doc Fireworks</a>
    
    % Author: Jiro Doke
    % Date: Dec 28, 2010
    % Copyright 2010 The MathWorks, Inc.

    %% Properties
    properties (SetObservable)
        % FireworksType  Type of fireworks
        %   String value that specifies the type of fireworks. To see a
        %   complete list of valid types, type
        %       <a href="matlab:Fireworks.ValidFireworksType">Fireworks.ValidFireworksType</a>
        %   'plain' and 'glitter' are the standard-looking fireworks, and
        %   one of the two is selected by default.
        %
        % See also Fireworks.ValidFireworksType
        FireworksType
        
        % NumPetals   Number of petals after explosion
        %   This is the number of spreading lines (petals) after the
        %   explosion at the top. The default is an integer randomly
        %   selected between 8 and 12.
        NumPetals
                
        % SymmetricPetals   Whether the petals spread symmetrically
        %   true/false. Default is randomly selected. This determines
        %   whether the spreading petals are symmetrically shaped. This
        %   only applies for 'plain' or 'glitter' FireworksType.
        SymmetricPetals
        
        % LaunchPos   Horizontal launch location
        %   A scalar number indicating the horizontal location of the
        %   launch. Default is a random value within the x limits of the
        %   axes.
        LaunchPos
        
        % LaunchVel   Launch velocity (horizontal and vertical)
        %   A two-element row vector indicating the horizontal and vertical
        %   velocity. Default is a pair of random values within a specific
        %   range.
        LaunchVel
        
        % SpeedFactor   Speed of the fireworks
        %   A scalar number indicating the speed factor. The value must be
        %   between 0.5 and 1.25. A small value corresponds to faster
        %   flight. Default is a random value between 0.5 and 1.25.
        SpeedFactor

        % HaveSound   Whether there is sound
        %   true/false. Default is true. This indicates whether there will
        %   be sound or not.
        HaveSound = true

        % ThinLines   Whether the fireworks are displayed with thin lines
        %   true/false. Default is false. This indicates whether to draw
        %   the fireworks with thin lines or not. Typically, this should be
        %   false. If you're displaying this in a very small axes, you may
        %   want to set this to true.
        ThinLines = false
    end
    
    properties (Dependent, GetAccess = protected, Hidden)
        LineWidth
        PetalColorIndex
    end
    
    properties (Constant)
        % ValidFireworksType    Valid strings for FireworksType property
        %
        %    'plain', 'glitter', 'heart', 'matlab', 'a', 'b', 'c', 'd',
        %    'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p',
        %    'q', 'r', 's', 't', 'u', 'v', 'w', 'x' 'y', 'z', '?', '!'
        %
        % See also Fireworks.FireworksType
        ValidFireworksType = [{'plain', 'glitter', 'heart', 'matlab'}, ...
            cellstr(('a':'z')')', {'?', '!'}]
    end
    
    properties (Constant, Hidden)
        Colormap = unique(getColorTable(), 'rows')
    end
    
    properties (Constant, GetAccess = protected, Hidden)
        ColorTable = getColorTable();
        Sounds = load('fireworks_sound')
    end
    
    properties (Access = public, Hidden)
        T_Stem
        Y_Stem
        T_Petals
        Y_Petals
        H_Stem
        H_Petals
        
        SwooshSound
        PopSound
        CrackleSound
        
        StartTime
        
        TailLength = 20
        PetalTailTime = 10
        
        Stage
        
        GlitterVals = [30 10];
        GlitterCount
        
        PropChangeListener
        
        axH
    end
    
    %% Public Methods
    methods
        %% Main Methods
        function obj = Fireworks(axH, varargin)
            % Fireworks   Creates a Fireworks object
            %
            %   FWOBJ = Fireworks(AX) creates a single fireworks object
            %       attached to axes AX. The properties of the object are
            %       randomly set.
            %   FWOBJ = Fireworks(AX, NUM) creates NUM fireworks at the
            %       same time. The properties of the objects are randomly
            %       set.
            %
            %   FWOBJ = Fireworks(AX,  PROP1, VAL1, PROP2, VAL2, ...)
            %   FWOBJ = Fireworks(AX, NUM,  PROP1, VAL1, PROP2, VAL2, ...)
            %       Pass in additional Property-Value pairs to fix those
            %       values. Properties not specified will be randomly
            %       set.
            %
            %   See the help for Fireworks class to learn more about valid
            %   properties.
            %
            % See also Fireworks, Fireworks.step, Fireworks.step,
            % Fireworks.flyBlocking, Fireworks.newFireworks,
            % Fireworks.resetFireworks
            
            % Calling with zero input arguments is used internally only.
            if nargin == 0
                return;
            end
                        
            validateattributes(axH, {'double'}, {'scalar'}, mfilename, 'axH');
            assert(ishandle(axH), 'Fireworks:Constructor:InvalidAxes', ...
                'The first argument must be a valid axes handle');
            
            if mod(length(varargin), 2) == 1
                validateattributes(varargin{1}, {'numeric'}, ...
                    {'scalar', 'positive', 'integer'}, mfilename, ...
                    'NumFireworks');
                numFireworks = varargin{1};
                varargin(1) = '';
            else
                numFireworks = 1;
            end
                        
            if numFireworks > 1
                for id = numFireworks:-1:1
                    obj(id) = Fireworks(axH, varargin{:});
                end
                return;
            end
            
            % Set up the axes (if it has not been already)
            if isempty(getappdata(axH, 'ReadyForFireworks'))
                oldUnits = get(axH, 'Units');
                set(axH, 'Units', 'pixels');
                pos = get(axH, 'Position');
                set(axH, ...
                    'Color', 'black', ...
                    'Units', oldUnits, ...
                    'YLim', [0 3000], ...
                    'XTick', [], ...
                    'YTick', [], ...
                    'Box', 'on', ...
                    'DataAspectRatio', [1 1 1], ...
                    'PlotBoxAspectRatio', [1 pos(4)/pos(3) 1]);
                set(axH, 'XLim', get(axH, 'XLim'));
                setappdata(axH, 'ReadyForFireworks', true)
            end
            
            obj.axH = axH;
            
            % Set up listeners for property change
            obj.PropChangeListener = addlistener(obj, ...
                {'FireworksType', 'SpeedFactor', 'NumPetals', ...
                'LaunchPos', 'LaunchVel', 'SymmetricPetals', ...
                'ThinLines'}, 'PostSet', ...
                @obj.propModifiedFcn);
            
            newFireworks(obj, varargin{:})
            
        end
                    
        function success = step(obj)
            % step   Step fireworks animation
            %
            %   step(FWOBJ) steps the fireworks animation. Note that this
            %       is time-based animation, not frame based. For each
            %       "step" it calculates the current position based on
            %       time. For this reason, you should run this in a loop or
            %       using a timer object. OBJ can be an array of Fireworks
            %       objects, and it would step each object.
            %
            %   S = step(FWOBJ) returns true if the procedure was
            %       successful. It would return false if any of the
            %       following occurs:
            %           1. The fireworks completed.
            %           2. The object was deleted.
            %           3. The axes was deleted.
            %       S will be the same size as OBJ.
            %
            %   To automatically play the whole animation, use the method
            %   "flyBlocking". Note that this will block MATLAB until
            %   finished.
            %
            %   Once the fireworks has completed, in order to launch it
            %   again, you need to call the "resetFireworks" method before
            %   you can step.
            %
            % Example:
            %   % Create 10 fireworks
            %       f = Fireworks(gca, 10);
            %
            %   % Animate at 0.1 second intervals and continue until all
            %   % have finished
            %       while any(step(f))
            %           pause(0.1);
            %       end
            %
            % See also Fireworks, Fireworks.Fireworks,
            % Fireworks.flyBlocking, Fireworks.resetFireworks,
            % Fireworks.newFireworks.
            
            % Treat method call to an array of objects
            if numel(obj) > 1
                success = false(size(obj));
                for id = 1:numel(obj)
                    success(id) = step(obj(id));
                end
                return;
            end
            
            if ~isOperable(obj)
                success = false;
                return;
            end
                
            if ~ishandle(obj.axH)
                delete(obj);
                success = false;
                return;
            end
            
            if isempty(obj.Stage)
                computeTrajectory(obj)
            end
            
            success = true;
            
            switch obj.Stage
                case -1
                    success = false;
                case 0
                    if obj.HaveSound
                        %play(obj.SwooshSound);
                    end
                    obj.StartTime = clock;
                    obj.GlitterCount = 1;
                    obj.Stage = 1;
                case {1, 2, 3}
                    et = etime(clock, obj.StartTime)*12;
                    switch obj.Stage
                        case 1
                            et = et/obj.SpeedFactor;
                            tailLength = 5;
                            if et < obj.T_Stem(end)
                                [minVal, id] = min(abs(obj.T_Stem-et));
                                set(obj.H_Stem, ...
                                    'XData', obj.Y_Stem(max(1, id-tailLength):id, 1), ...
                                    'YData', obj.Y_Stem(max(1, id-tailLength):id, 3));
                            else
                                set(obj.H_Stem, 'XData', NaN, 'YData', NaN);
                                obj.Stage = 2;
                                if obj.HaveSound
                                    play(obj.PopSound);
                                end
                            end
                        case 2
                            et = et - obj.T_Stem(end)*obj.SpeedFactor;
                            if et < obj.T_Petals(end)
                                [minVal, id] = min(abs(obj.T_Petals-et));
                                currentTailLength = round(id/length(obj.T_Petals)*obj.TailLength);
                                sID = max(1, id-currentTailLength);
                                eID = min(length(obj.T_Petals), id);
                                idx = sID:eID;
                                set(obj.H_Petals, {'XData', 'YData'}, ...
                                    [...
                                    mat2cell(obj.Y_Petals(idx, 1:4:end)', ones(1, obj.NumPetals), length(idx)), ...
                                    mat2cell(obj.Y_Petals(idx, 3:4:end)', ones(1, obj.NumPetals), length(idx))]);
                            elseif et <= obj.T_Petals(end) + obj.PetalTailTime
                                pct = 1 - (et - obj.T_Petals(end))/obj.PetalTailTime;
                                sID = length(obj.T_Petals) - round(obj.TailLength*pct);
                                eID = length(obj.T_Petals);
                                idx = sID:eID;
                                c = obj.getColor(obj.PetalColorIndex, 1-pct);
                                set(obj.H_Petals, {'Color', 'XData', 'YData'}, ...
                                    [...
                                    mat2cell(c, ones(1, obj.NumPetals), 3), ...
                                    mat2cell(obj.Y_Petals(idx, 1:4:end)', ones(1, obj.NumPetals), length(idx)), ...
                                    mat2cell(obj.Y_Petals(idx, 3:4:end)', ones(1, obj.NumPetals), length(idx))]);
                            else
                                if strcmp(obj.FireworksType, 'plain')
                                    set(obj.H_Petals, 'XData', NaN, 'YData', NaN);
                                    obj.Stage = -1;
                                    success = false;
                                else
                                    obj.Stage = 3;
                                end
                            end
                        case 3
                            et2 = et - (obj.T_Stem(end)*obj.SpeedFactor + obj.T_Petals(end) + obj.PetalTailTime);
                            if obj.GlitterCount == 1
                                if obj.HaveSound
                                    play(obj.CrackleSound);
                                end
                                
                                % First time through, randomly choose points around
                                % each tip
                                set(obj.H_Petals, ...
                                    'LineStyle', 'none', ...
                                    'LineWidth', 0.5, ...
                                    'Marker', 'o', ...
                                    {'XData', 'YData'}, ...
                                    [...
                                    mat2cell(bsxfun(@plus, obj.GlitterVals(1)*randn(obj.NumPetals, 4), obj.Y_Petals(end, 1:4:end)'), ones(1, obj.NumPetals), 4), ...
                                    mat2cell(bsxfun(@plus, obj.GlitterVals(1)*randn(obj.NumPetals, 4), obj.Y_Petals(end, 3:4:end)'), ones(1, obj.NumPetals), 4)]);
                                
                                obj.GlitterCount = 1.5;
                                
                            elseif ceil(et2/1) > obj.GlitterCount && et2 <= 30
                                % Update only every 1 count (or so)
                                
                                % For subsequent calls, tweak the last location by
                                % some small amount. Also make it fall (by 10
                                % points) to account for gravity (YData)
                                obj.GlitterCount = ceil(et2/1);
                                
                                c1 = obj.getColor(obj.PetalColorIndex, et2/30/1);
                                c2 = obj.getColor(obj.PetalColorIndex, et2/30*2);
                                
                                d = get(obj.H_Petals, {'XData', 'YData'});
                                d(:, 1) = mat2cell(cell2mat(d(:, 1))+obj.GlitterVals(2)*randn(obj.NumPetals, 4), ...
                                    ones(1, obj.NumPetals), 4);
                                d(:, 2) = mat2cell(cell2mat(d(:, 2))+obj.GlitterVals(2)*randn(obj.NumPetals, 4)-10, ...
                                    ones(1, obj.NumPetals), 4);
                                c1_cell = mat2cell(c1, ones(1, obj.NumPetals), 3);
                                c2_cell = mat2cell(c2, ones(1, obj.NumPetals), 3);
                                set(obj.H_Petals, ...
                                    {'XData', 'YData', 'MarkerFaceColor', 'MarkerEdgeColor'}, ...
                                    [d, c1_cell, c2_cell]);
                                
                            elseif et2 > 30
                                set(obj.H_Petals, 'XData', NaN, 'YData', NaN);
                                obj.Stage = -1;
                                success = false;
                            end
                    end
            end
            % drawnow;
                        
        end
        
        function flyBlocking(obj)
            % flyBlocking   Fly fireworks to completion (blocking mode)
            %
            %   flyBlocking(FWOBJ) animates the fireworks in its entirety.
            %       This blocks MATLAB until all animation has completed.
            %       FWOBJ can be an array of fireworks objects. It calls
            %       the "resetFireworks" method in the beginning to ensure
            %       that all fireworks are reset to their initial
            %       condition.
            %
            %   This method is a convenience function for the following set
            %   of commands:
            %
            %         resetFireworks(obj);
            %         while any(step(obj))
            %             drawnow;
            %         end
            %
            % See also Fireworks, Fireworks.Fireworks, Fireworks.step,
            % Fireworks.resetFireworks, Fireworks.newFireworks.
            
            resetFireworks(obj);
            while any(step(obj))
                drawnow;
            end
        end
        
        function newFireworks(obj, varargin)
            % newFireworks   Change fireworks properties
            %
            %   newFireworks(FWOBJ) redefines the properties of the
            %       fireworks object(s) to their randomly set values. FWOBJ
            %       can be an array of Fireworks objects.
            %
            %   newFireworks(FWOBJ, PROP1, VAL1, PROP2, VAL2, ...) sets
            %       the specified properties, while randomly setting the
            %       non-specified properties. PROPs can be any of the valid
            %       properties of the Fireworks class. For an array FWOBJ,
            %       the specified properties are common across all
            %       Fireworks objects, and the non-specified properties are
            %       randomly set.
            %
            % Example:
            %   % Create 10 fireworks and animate
            %      f2 = Fireworks(gca, 10);
            %      flyBlocking(f2);
            %
            %   % Change the LaunchPos property of all 10 fireworks to 0.
            %   % The other properties are randomly chosen.
            %      newFireworks(f2, 'LaunchPos', 0);
            %      flyBlocking(f2);
            %
            % See also Fireworks, Fireworks.Fireworks, Fireworks.step,
            % Fireworks.flyBlocking, Fireworks.resetFireworks.
            
            for id = 1:length(obj)
                if isOperable(obj(id))
                    xl = xlim80(obj(id));
                    
                    % Random default values
                    obj(id).FireworksType   = Fireworks.ValidFireworksType{round(rand)+1};
                    obj(id).SpeedFactor     = 0.75*rand + 0.5;
                    obj(id).NumPetals       = 8 + round(4*rand);
                    obj(id).LaunchPos       = diff(xl)*rand+xl(1);
                    obj(id).LaunchVel       = [80*rand-40, 200+(60*rand-30)];
                    obj(id).SymmetricPetals = logical(round(rand));
                    
                    for id2 = 1:length(varargin)/2
                        prop = validatestring(varargin{id2*2-1}, ...
                            {'FireworksType', 'SpeedFactor', 'NumPetals', ...
                            'LaunchPos', 'LaunchVel', 'SymmetricPetals', ...
                            'ThinLines', 'HaveSound'}, mfilename, 'PropertyName');
                        obj(id).(prop) = varargin{id2*2};
                    end
                    
                    computeTrajectory(obj(id));
                end
            end
                        
        end
                
        function resetFireworks(obj)
            % resetFireworks    Reset animation to intial condition
            %
            %   resetFireworks(FWOBJ) resets the animation of the Fireworks
            %       object(s) to their respective initial conditions. This
            %       is required if you want to re-animate the same
            %       fireworks once it has completed. There are few
            %       situations where resetting happens automatically:
            %           1. At the start of "flyBlocking" method
            %           2. At the end of "newFireworks" method
            %           3. After modifying any of the properties
            %
            % See also Fireworks, Fireworks.Fireworks, Fireworks.step,
            % Fireworks.flyBlocking, Fireworks.newFireworks.
            
            for id = 1:length(obj)
                if isOperable(obj(id))
                    obj(id).Stage = 0;
                    c = getColor(obj(id), obj(id).PetalColorIndex, 0);
                    set(obj(id).H_Petals, ...
                        'LineStyle', '-', ...
                        'LineWidth', obj(id).LineWidth, ...
                        'Marker', 'none', ...
                        'XData', NaN, ...
                        'YData', NaN, ...
                        {'Color'}, mat2cell(c, ones(1, obj(id).NumPetals), 3));
                    set(obj(id).H_Stem, 'XData', NaN, 'YData', NaN);
                end
            end
        end
        
        %% Getter Functions
        function val = get.PetalColorIndex(obj)
            val = round(7*rand(1, obj.NumPetals))+1;
        end
        
        function val = get.LineWidth(obj)
            if obj.ThinLines
                val = 0.5;
            else
                val = 2;
            end
        end
        
        %% Setter Functions
        function set.FireworksType(obj, val)
            val = validatestring(val, Fireworks.ValidFireworksType, ...
                mfilename, 'FireworksType');
            obj.FireworksType = val;
        end
        
        function set.SpeedFactor(obj, val)
            validateattributes(val, {'numeric'}, {'real', 'scalar'}, ...
                mfilename, 'SpeedFactor');
            assert(val >= 0.5 && val <= 1.25, ...
                'Fireworks:SpeedFactor:InvalidRange', ...
                'SpeedFactor must be a scalar between 0.5 and 1.25');
            obj.SpeedFactor = val;
        end
        
        function set.NumPetals(obj, val)
            validateattributes(val, {'numeric'}, {'real', 'scalar'}, ...
                mfilename, 'NumPetals');
            assert(val >= 8 && val <= 20, ...
                'Fireworks:NumPetals:InvalidRange', ...
                'NumPetals must be a scalar between 8 and 20');
            obj.NumPetals = val;
        end
        
        function set.HaveSound(obj, val)
            validateattributes(val, {'logical'}, {'scalar'}, ...
                mfilename, 'HaveSound');
            obj.HaveSound = val;
        end
        
        function set.LaunchPos(obj, val)
            if isOperable(obj)                
                xl = xlim(obj);
                validateattributes(val, {'numeric'}, {'real', 'scalar'}, ...
                    mfilename, 'LaunchPos');
                assert(val >= xl(1) && val <= xl(2), ...
                    'Fireworks:LaunchPos:InvalidRange', ...
                    'LaunchPos must be a scalar between %d and %d', xl(1), xl(2));
                obj.LaunchPos = val;
            end
        end
        
        function set.LaunchVel(obj, val)
            validateattributes(val, {'numeric'}, {'real', 'vector'}, ...
                mfilename, 'LaunchVel');
            assert(isequal(size(val), [1 2]) && all(val >= -1000) && ...
                all(val <= 1000), 'Fireworks:LaunchVel:InvalidRange', ...
                'LaunchVel must be a 2-element vector between -1000 and 1000');
            obj.LaunchVel = val;
        end
        
        function set.SymmetricPetals(obj, val)
            validateattributes(val, {'logical'}, {'scalar'}, ...
                mfilename, 'SymmetricPetals');
            obj.SymmetricPetals = val;
        end
        
        function set.ThinLines(obj, val)
            validateattributes(val, {'logical'}, {'scalar'}, ...
                mfilename, 'ThinLines');
            obj.ThinLines = val;
        end
        
    end
    
    %% Protected Methods
    methods (Access = protected, Hidden)
        
        function tf = isOperable(obj)
           tf = isvalid(obj) && ~isempty(obj.axH); 
        end
        
        function tf = isCustomFireworks(obj)
            tf = ~ismember(obj.FireworksType, {'plain', 'glitter'});
        end

        function propModifiedFcn(obj, varargin)
            obj.Stage = [];
        end
        
        function computeTrajectory(obj)
            if ~isOperable(obj)
                return;
            end
            
            if ~ishandle(obj.axH)
               delete(obj);
               return;
            end
                        
            m = 1;
            b = 0;
            
            % Initial conditions
            y0 = [obj.LaunchPos, obj.LaunchVel(1), 0, obj.LaunchVel(2)];
            
            % Stop at the top of the flight
            opt = odeset('Events', @terminateFcn);
            [T, Y] = ode45(@(t, y) odeFcn(t, y, m, b), ...
                [0 100], y0, opt);
            
            % Re-solve with finer time steps
            [obj.T_Stem, obj.Y_Stem] = ode45(@(t, y) odeFcn(t, y, m, b), ...
                linspace(0, T(end)), y0);
            
            if isCustomFireworks(obj)
                [vel, ang] = getCustomVelAng(obj.FireworksType);
                obj.NumPetals = length(vel);
                obj.GlitterVals = [20 5];
                obj.TailLength = 5;
                obj.PetalTailTime = 5;
            else
                obj.GlitterVals = [30 10];
                obj.TailLength = 20;
                obj.PetalTailTime = 10;
                vel = 35;
            end
            obj.T_Petals = linspace(0, 12, 50)';
            obj.Y_Petals = nan(50, 4*obj.NumPetals);
            g = 2;
            
            for id = 1:obj.NumPetals
                if isCustomFireworks(obj)
                    [T2, obj.Y_Petals(:, id*4-3:id*4)] = ode45(@(t, y) odeFcn(t, y, m, b, g), ...
                        obj.T_Petals, [obj.Y_Stem(end, 1), vel(id)*cos(ang(id)), obj.Y_Stem(end, 3), vel(id)*sin(ang(id))]);
                    
                else
                    ang = id*2*pi/obj.NumPetals + 0.1*rand-0.5;
                    if obj.SymmetricPetals
                        [T2, obj.Y_Petals(:, id*4-3:id*4)] = ode45(@(t, y) odeFcn(t, y, m, b, g), ...
                            obj.T_Petals, [obj.Y_Stem(end, 1), vel*cos(ang), obj.Y_Stem(end, 3), vel*sin(ang)]);
                    else
                        [T2, obj.Y_Petals(:, id*4-3:id*4)] = ode45(@(t, y) odeFcn(t, y, m, b, g), ...
                            obj.T_Petals, [obj.Y_Stem(end, 1), (vel+20*rand-10)*cos(ang), obj.Y_Stem(end, 3), (vel+20*rand-10)*sin(ang)]);
                    end
                end
            end
            
            speed_breaks = [0 15 20 25 inf];
            obj.SwooshSound = loadSound('swoosh');
            obj.PopSound = loadSound('pop', ...
                6 - find(speed_breaks > T(end)* obj.SpeedFactor, 1));
            obj.CrackleSound = loadSound('crackle');
            
            % Prepare plot
            if ishandle(obj.H_Stem)
                delete(obj.H_Stem);
            end
            if ishandle(obj.H_Petals)
                delete(obj.H_Petals)
            end
            
            % Stem
            obj.H_Stem = line(obj.Y_Stem(1, 1), obj.Y_Stem(1, 3), ...
                'Parent', obj.axH, ...
                'LineWidth', obj.LineWidth, ...
                'Color', [1 1 1]);
            
            % Petals
            obj.H_Petals = nan(1, obj.NumPetals);
            c = obj.getColor(obj.PetalColorIndex, 0);
            for id = 1:obj.NumPetals
                obj.H_Petals(id) = line(nan, nan, ...
                    'Parent', obj.axH, ...
                    'Color', c(id, :), ...
                    'LineWidth', obj.LineWidth, ...
                    'MarkerSize', obj.LineWidth, ...
                    'MarkerFaceColor', c(id, :), ...
                    'MarkerEdgeColor', c(id, :));
            end
            
            resetFireworks(obj);
        end

        function val = xlim(obj)
            val = xlim(obj.axH);
        end

        function val = xlim80(obj)
            xl = xlim(obj.axH);
            val = xl + [1, -1]*diff(xl)*0.1;
        end
        
        function color = getColor(obj, id, pct)
            color = obj.ColorTable((id-1)*32+2+round(min(pct, 1)*31), :);
        end
                
    end
    
    methods (Hidden)

        function delete(obj)
            delete@handle(obj);
            
            if ishandle(obj.H_Stem), delete(obj.H_Stem), end
            if ishandle(obj.H_Petals), delete(obj.H_Petals), end
        end
 
    end
    
    %% Hidden Methods (overloading handle methods)
    methods (Hidden)
        
        function out = addlistener(varargin)
            out = addlistener@handle(varargin{:});
        end
        
        function notify(varargin)
            notify@handle(varargin{:})
        end
        
        function out = eq(varargin)
            out = eq@handle(varargin{:});
        end
        
        function out = ge(varargin)
            out = ge@handle(varargin{:});
        end
        
        function out = gt(varargin)
            out = gt@handle(varargin{:});
        end
        
        function out = le(varargin)
            out = le@handle(varargin{:});
        end
        
        function out = lt(varargin)
            out = lt@handle(varargin{:});
        end
        
        function out = ne(varargin)
            out = ne@handle(varargin{:});
        end
        
        function out = findprop(varargin)
            out = findprop@handle(varargin{:});
        end
        
        function findobj(varargin)
            findobj@handle(varargin{:});
        end
                
    end
    
    
end

%#ok<*INUSL>
%#ok<*NASGU>
%#ok<*ASGLU>